@@ -14,12 +14,12 @@ def top_p(logits, thres = 0.9):
     sorted_logits, sorted_indices = torch.sort(logits, descending=True)
     cum_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)
 
-    sorted_indices_to_remove = cum_probs > thres
+    sorted_indices_to_remove = cum_probs > 1.0 - thres
     sorted_indices_to_remove[:, 1:] = sorted_indices_to_remove[:, :-1].clone()
     sorted_indices_to_remove[:, 0] = 0
 
     sorted_logits[sorted_indices_to_remove] = float('-inf')
-    return sorted_logits.gather(1, sorted_indices)
+    return sorted_logits.scatter(1, sorted_indices, sorted_logits)
 
 def top_k(logits, thres = 0.9):
     k = int((1 - thres) * logits.shape[-1])
